//
//  SpikeCamera.m
//  SpikeNav
//
//  Created by Robert Bridson on 24/06/09.
//  Released into the public domain.
//

#import "SpikeCamera.h"
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>

// a smooth function for -1<=t<=1, where f(t0)=0, f'(t0)=s0,
// and f(-1)=fneg, f(1)=fpos.
static float accel_ramp(float t, float t0, float s0, float fneg, float fpos)
{
   if(t<t0){
      return s0*(t-t0) + (s0*(t0+1)+fneg)/((t0+1)*(t0+1))*(t-t0)*(t-t0);
   }else if(t>t0){
      return s0*(t-t0) + (s0*(t0-1)+fpos)/((t0-1)*(t0-1))*(t-t0)*(t-t0);
   }else
      return 0;
}

// ------------------------------------------------------------------------------------------------------------------

@implementation SpikeCamera

- (SpikeCamera*) initWithRadius: (float) radius
{
   [super init];
   camera_position[0]=0;
   camera_position[1]=0.5*radius;
   camera_position[2]=3*radius;
   heading=0;
   pitch=0;
   perspective=0.2;
   near=0.01*radius;
   far=5*radius;
   width=1;
   height=1;
   aspect=1;
   lever_threshold=0.95;
   lever_min=0.2;
   anchor_camz=radius;
   return self;
}

- (void) prepareToRender: (NSRect) pixel_bounds withRadius: (float) radius
{
   // grab image dimensions and set the viewport
   width=pixel_bounds.size.width;
   height=pixel_bounds.size.height;
   glViewport(0, 0, width, height);
   
   // clear the screen in preparation for drawing
   glClearColor(1, 1, 1, 0);
   glClearDepth(1);
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glEnable(GL_DEPTH_TEST);
   glDepthMask(GL_TRUE);
   
   // set up projection
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   aspect=width/height;
   near=0.01*radius;
   far=radius+sqrt(camera_position[0]*camera_position[0]
                   +camera_position[1]*camera_position[1]
                   +camera_position[2]*camera_position[2]);
   glFrustum(-perspective*aspect*near, perspective*aspect*near,
             -perspective*near, perspective*near,
             near, far);
   
   // set up modelview
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   glTranslatef(-camera_position[0], -camera_position[1], -camera_position[2]);
   glRotatef(pitch, 1, 0, 0);
   glRotatef(heading, 0, 1, 0);
}

- (void) renderAdditions
{
   // draw anchor
   glDepthMask(GL_FALSE);
   glLineWidth(3);
   glBegin(GL_LINES);
   glDisable(GL_LIGHTING);
   glColor4f(1, 0, 0, 0.5);
   glVertex3f(anchor_worldspace[0]-5*near, anchor_worldspace[1], anchor_worldspace[2]);
   glVertex3f(anchor_worldspace[0]+5*near, anchor_worldspace[1], anchor_worldspace[2]);
   glColor4f(0, 1, 0, 0.5);
   glVertex3f(anchor_worldspace[0], anchor_worldspace[1]-5*near, anchor_worldspace[2]);
   glVertex3f(anchor_worldspace[0], anchor_worldspace[1]+5*near, anchor_worldspace[2]);
   glColor4f(0, 0, 1, 0.5);
   glVertex3f(anchor_worldspace[0], anchor_worldspace[1], anchor_worldspace[2]-5*near);
   glVertex3f(anchor_worldspace[0], anchor_worldspace[1], anchor_worldspace[2]+5*near);
   glEnd();
   glDepthMask(GL_TRUE);
}

- (void) mouseDown: (NSEvent*) event atPoint: (NSPoint) local_point;
{
   // convert this to a 3d world space point
   // and save for future reference as "anchor"
   [self unproject:local_point toWorldSpace:anchor_worldspace andCamZ:&anchor_camz];
   // if we're dealing with a tablet, check those stylus parameters
   if([event subtype]==NSTabletPointEventSubtype){
      lever=([event pressure]>lever_threshold);
      old_tilt=[event tilt];
      old_heading=heading;
      old_pitch=pitch;
      old_camz=anchor_camz;
   }      
}

- (void) mouseDragged: (NSEvent*) event atPoint: (NSPoint) local_point
{
   // if tablet, do extra work
   if([event subtype]==NSTabletPointEventSubtype){
      if([event pressure]>lever_threshold){
         if(lever==NO){
            old_tilt=[event tilt]; // treat it as a new stylus down for mdoe switch
            old_heading=heading;
            old_pitch=pitch;
            old_camz=anchor_camz;
            lever=YES;
         }
      }else if([event pressure]<lever_min){
         if(lever==YES){
            old_tilt=[event tilt]; // treat it as a new stylus down for mode switch
            old_heading=heading;
            old_pitch=pitch;
            old_camz=anchor_camz;
            lever=NO;
         }
      }
      NSPoint new_tilt=[event tilt];
      if(lever){ // dollying
         float delta=new_tilt.y-old_tilt.y;
         anchor_camz=old_camz*exp(-2.5*delta);
      }else{ // rotating
         heading=old_heading+accel_ramp(new_tilt.x, old_tilt.x, 60, -210, 210);
         if(heading<180) heading+=360;
         else if(heading>180) heading-=360;
         
         pitch=old_pitch+accel_ramp(-new_tilt.y, -old_tilt.y, 60, -180, 180);
         if(pitch>90) pitch=90;
         else if(pitch<-90) pitch=-90;
      }
   }
   // adjust camera to translate anchor as desired
   [self solveCameraPositionForBasePoint:local_point];
}

- (void) mouseUp: (NSEvent*) event atPoint: (NSPoint) local_point
{
   [self mouseDragged:event atPoint: local_point];
}

- (void) unproject: (NSPoint) base_point toWorldSpace: (float[3]) world_space andCamZ: (float*) camz
{
   // first grab raw z from depth buffer
   glReadBuffer(GL_FRONT);
   float depth;
   glReadPixels(base_point.x, base_point.y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &depth);   
   // calculate camera coordinates
   if(depth==1) *camz=anchor_camz; // don't use far clipping plane if we clicked on background
   else *camz=2*far*near/((2*depth-1)*(far-near)-(far+near));
   float ndcx=2*base_point.x/width-1, ndcy=2*base_point.y/height-1;
   float camx=-ndcx*perspective*aspect*(*camz);
   float camy=-ndcy*perspective*(*camz);
   // then calculate world space position
   float r=(camx+camera_position[0]), s=(camy+camera_position[1]), t=(*camz+camera_position[2]);
   float ch=cos((2*M_PI/360)*heading), sh=sin((2*M_PI/360)*heading);
   float cp=cos((2*M_PI/360)*pitch), sp=sin((2*M_PI/360)*pitch);
   world_space[0]=ch*r-sh*(-sp*s+cp*t);
   world_space[1]=cp*s+sp*t;
   world_space[2]=sh*r+ch*(-sp*s+cp*t);
   /*
    GLdouble mv[16], proj[16], x, y, z;
    GLint vp[4];
    glGetIntegerv(GL_VIEWPORT, vp);
    glGetDoublev(GL_MODELVIEW_MATRIX, mv);
    glGetDoublev(GL_PROJECTION_MATRIX, proj);
    gluUnProject(base_point.x, base_point.y, depth, mv, proj, vp, &x, &y, &z);
    world_space[0]=x;
    world_space[1]=y;
    world_space[2]=z;
    */
}

- (void) solveCameraPositionForBasePoint: (NSPoint) base_point
{
   // figure out camera_position that causes anchor_worldspace to project to base_point and
   // also get mapped to anchor_camz.
   
   // together with anchor_camz, these are the desired camera coordinates of the anchor:
   float ndcx=2*base_point.x/width-1, ndcy=2*base_point.y/height-1;
   float anchor_camx=-ndcx*perspective*aspect*(anchor_camz);
   float anchor_camy=-ndcy*perspective*(anchor_camz);
   
   // rotate the anchor
   float ch=cos((2*M_PI/360)*heading), sh=sin((2*M_PI/360)*heading);
   float cp=cos((2*M_PI/360)*pitch), sp=sin((2*M_PI/360)*pitch);
   float anchor_rotx=ch*anchor_worldspace[0]+sh*anchor_worldspace[2];
   float anchor_roty=cp*anchor_worldspace[1]-sp*(-sh*anchor_worldspace[0]+ch*anchor_worldspace[2]);
   float anchor_rotz=sp*anchor_worldspace[1]+cp*(-sh*anchor_worldspace[0]+ch*anchor_worldspace[2]);
   
   // and find the difference to figure out the camera
   camera_position[0]=anchor_rotx-anchor_camx;
   camera_position[1]=anchor_roty-anchor_camy;
   camera_position[2]=anchor_rotz-anchor_camz;
}

@end
